"""AST comprehension classes and functions."""

from __future__ import annotations

from abc import abstractmethod
from typing import (
    Iterable,
    Optional,
)

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    Expr,
    ReprStruct,
    SourceLocation,
)
from astx.tools.typing import typechecked


@public
@typechecked
class ComprehensionClause(Expr):
    """AST node for generic comprehensions."""

    target: Expr
    iterable: Expr
    conditions: ASTNodes[Expr]
    is_async: bool

    def __init__(
        self,
        target: Expr,
        iterable: Expr,
        conditions: Optional[Iterable[Expr] | ASTNodes[Expr]] = None,
        is_async: bool = False,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)
        self.target = target
        self.iterable = iterable
        self.is_async = is_async
        self.kind = ASTKind.ComprehensionKind

        if isinstance(conditions, ASTNodes):
            self.conditions = conditions
        elif isinstance(conditions, Iterable):
            self.conditions = ASTNodes()
            for condition in conditions:
                self.conditions.append(condition)
        else:
            self.conditions = ASTNodes[Expr]()

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"COMPREHENSION[is_async={self.is_async}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        conditions = (
            {"conditions": self.conditions.get_struct(simplified)}
            if self.conditions.nodes
            else {}
        )

        value: ReprStruct = {
            "target": self.target.get_struct(simplified),
            "iterable": self.iterable.get_struct(simplified),
            **conditions,
        }

        key = f"{self}" if not simplified else f"{self}#{id(self)}"
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class Comprehension(Expr):
    """AST Comprehension class."""

    generators: ASTNodes[ComprehensionClause]

    def __init__(
        self,
        generators: (
            Iterable[ComprehensionClause] | ASTNodes[ComprehensionClause]
        ),
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)

        if isinstance(generators, ASTNodes):
            self.generators = generators
        elif isinstance(generators, Iterable):
            self.generators = ASTNodes[ComprehensionClause]()
            for generator in generators:
                self.generators.append(generator)

    @abstractmethod
    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        value: ReprStruct = {
            "generators": self.generators.get_struct(simplified),
        }
        key = f"{self}" if not simplified else f"{self}#{id(self)}"
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ListComprehension(Comprehension):
    """ListComprehension class."""

    element: Expr

    def __init__(
        self,
        element: Expr,
        generators: ASTNodes[ComprehensionClause]
        | Iterable[ComprehensionClause] = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the GeneratorExpr instance."""
        super().__init__(generators=generators, loc=loc, parent=parent)
        self.element = element

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"{self}"
        key += f"#{id(self)}" if simplified else ""

        generators = (
            {"generators": self.generators.get_struct(simplified)}
            if self.generators.nodes
            else {}
        )

        value: ReprStruct = {
            "element": self.element.get_struct(simplified),
            **generators,
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class SetComprehension(Comprehension):
    """AST node representing set comprehension expressions."""

    element: Expr

    def __init__(
        self,
        element: Expr,
        generators: ASTNodes[ComprehensionClause]
        | Iterable[ComprehensionClause] = [],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the SetComprehension instance."""
        super().__init__(generators=generators, loc=loc, parent=parent)
        self.element = element
        self.kind = ASTKind.SetComprehensionKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return "SET-COMPREHENSION"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        generators = (
            {"generators": self.generators.get_struct(simplified)}
            if self.generators.nodes
            else {}
        )

        value: ReprStruct = {
            "element": self.element.get_struct(simplified),
            **generators,
        }
        key = f"{self}#{id(self)}" if simplified else f"{self}"
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class GeneratorExpr(Comprehension):
    """AST class for generator expressions."""

    element: Expr

    def __init__(
        self,
        element: Expr,
        generators: Iterable[ComprehensionClause]
        | ASTNodes[ComprehensionClause],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the GeneratorExpr instance."""
        super().__init__(generators=generators, loc=loc, parent=parent)
        self.element = element
        self.kind = ASTKind.GeneratorExprKind

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"GENERATOR-EXPR#{id(self)}" if simplified else "GENERATOR-EXPR"
        value: ReprStruct = {
            "element": self.element.get_struct(simplified),
            "generators": self.generators.get_struct(simplified),
        }
        return self._prepare_struct(key, value, simplified)



"""Modifications for ASTx, such as visibility, scope, etc."""

from enum import Enum

from public import public


@public
class VisibilityKind(Enum):
    """Definition of different kind of visibility."""

    public = 1
    private = 2
    protected = 3


@public
class ScopeKind(Enum):
    """Definition for different kind of scopes."""

    global_ = 1
    local = 2


@public
class MutabilityKind(Enum):
    """Definition for different kind of mutability."""

    constant = 1
    mutable = 2



"""ASTx classes for the operators."""

from __future__ import annotations

from typing import Iterable, Literal, Optional, cast

from public import public
from typing_extensions import TypeAlias

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DataType,
    DictDataTypesStruct,
    Expr,
    Identifier,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.tools.typing import typechecked
from astx.variables import Variable


@public
@typechecked
class WalrusOp(DataType):
    """AST class for the Walrus (assignment expression) operator."""

    def __init__(
        self,
        lhs: Variable,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the WalrusOp instance."""
        super().__init__(loc=loc)
        self.lhs = lhs
        self.rhs = rhs
        self.kind = ASTKind.WalrusOpKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"WalrusOp[:=]({self.lhs} := {self.rhs})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure that represents the object."""
        key = "WALRUS[:=]"
        lhs = {"lhs": self.lhs.get_struct(simplified)}
        rhs = {"rhs": self.rhs.get_struct(simplified)}

        content: ReprStruct = {**lhs, **rhs}
        return self._prepare_struct(key, content, simplified)


@public
@typechecked
class AssignmentExpr(Expr):
    """AST class for assignment expressions."""

    targets: ASTNodes[Expr]
    value: Expr

    def __init__(
        self,
        targets: Iterable[Expr] | ASTNodes[Expr],
        value: Expr,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(loc=loc, parent=parent)

        if isinstance(targets, ASTNodes):
            self.targets = targets
        else:
            self.targets = ASTNodes()
            for target in targets:
                self.targets.append(target)

        self.value = value
        self.kind = ASTKind.AssignmentExprKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"AssignmentExpr[{self.value}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = "ASSIGNMENT-EXPR"
        targets_dict = {"targets": self.targets.get_struct(simplified)}
        value_dict = {"value": self.value.get_struct(simplified)}

        value = {
            **cast(DictDataTypesStruct, targets_dict),
            **cast(DictDataTypesStruct, value_dict),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class VariableAssignment(StatementType):
    """AST class for variable declaration."""

    name: str
    value: Expr

    def __init__(
        self,
        name: str,
        value: Expr,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the VarExprAST instance."""
        super().__init__(loc=loc, parent=parent)
        self.loc = loc
        self.name = name
        self.value = value
        self.kind = ASTKind.VariableAssignmentKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"VariableAssignment[{self.name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = str(self)
        value = self.value.get_struct(simplified)
        return self._prepare_struct(key, value, simplified)


OpCodeAugAssign: TypeAlias = Literal[
    "+=",
    "-=",
    "*=",
    "/=",
    "//=",
    "%=",
    "**=",
    "&=",
    "|=",
    "^=",
    "<<=",
    ">>=",
]


@public
@typechecked
class AugAssign(DataType):
    """AST class for augmented assignment."""

    target: Identifier
    op_code: OpCodeAugAssign
    value: DataType

    def __init__(
        self,
        target: Identifier,
        op_code: OpCodeAugAssign,
        value: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        super().__init__(loc=loc)
        self.target = target
        self.op_code = op_code
        self.value = value
        self.kind = ASTKind.AugmentedAssignKind

    def __str__(self) -> str:
        """Return a string that represents the augmented assignment object."""
        return f"AugAssign[{self.op_code}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = str(self)
        value: ReprStruct = {
            "target": self.target.get_struct(simplified),
            "value": self.value.get_struct(simplified),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class CompareOp(DataType):
    """AST class for comparison operators acting as properties."""

    def __init__(
        self,
        left: DataType,
        ops: Iterable[Literal["==", "!=", "<", ">", "<=", ">="]],
        comparators: Iterable[DataType],
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the CompareOp instance."""
        super().__init__(loc=loc)
        self.ops = list(ops)
        self.comparators = list(comparators)
        if len(self.ops) != len(self.comparators):
            raise ValueError(
                "Number of operators must equal number of comparators."
            )
        for op in self.ops:
            if op not in ["==", "!=", "<", ">", "<=", ">="]:
                raise ValueError(f"Invalid comparison operator: {op}")
        self.left = left
        self.kind = ASTKind.CompareOpKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        ops_str = ", ".join(self.ops)
        return f"Compare[{ops_str}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure that represents the object."""
        ops_str = ", ".join(self.ops)
        key = f"COMPARE[{ops_str}]"
        content: ReprStruct = {
            "left": self.left.get_struct(simplified),
            "comparators": [
                comp.get_struct(simplified) for comp in self.comparators
            ],
        }
        return self._prepare_struct(key, content, simplified)



"""Constant objects."""

from __future__ import annotations

from astx.types.base import NoneType
from astx.types.boolean import Boolean
from astx.types.numeric import (
    Complex,
    Complex32,
    Complex64,
    Float16,
    Float32,
    Float64,
    Floating,
    Int8,
    Int16,
    Int32,
    Int64,
    Int128,
    Integer,
    Number,
    SignedInteger,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    UInt128,
    UnsignedInteger,
)
from astx.types.string import (
    String,
    UTF8Char,
    UTF8String,
)
from astx.types.temporal import (
    Date,
    DateTime,
    Temporal,
    Time,
    Timestamp,
)

t_none = NoneType()
t_boolean = Boolean()
t_number = Number()
t_integer = Integer()
t_unsigned_integer = UnsignedInteger()
t_signed_integer = SignedInteger()
t_int8 = Int8()
t_int16 = Int16()
t_int32 = Int32()
t_int64 = Int64()
t_int128 = Int128()
t_uint8 = UInt8()
t_uint16 = UInt16()
t_uint32 = UInt32()
t_uint64 = UInt64()
t_uint128 = UInt128()
t_floating = Floating()
t_float16 = Float16()
t_float32 = Float32()
t_float64 = Float64()
t_complex = Complex()
t_complex32 = Complex32()
t_complex64 = Complex64()
t_string = String()
t_utf8_string = UTF8String()
t_utf8_char = UTF8Char()
t_temporal = Temporal()
t_date = Date()
t_time = Time()
t_timestamp = Timestamp()
t_datetime = DateTime()

__all__ = [
    "t_boolean",
    "t_complex",
    "t_complex32",
    "t_complex64",
    "t_date",
    "t_datetime",
    "t_float16",
    "t_float32",
    "t_float64",
    "t_floating",
    "t_int8",
    "t_int16",
    "t_int32",
    "t_int64",
    "t_int128",
    "t_integer",
    "t_none",
    "t_number",
    "t_signed_integer",
    "t_string",
    "t_temporal",
    "t_time",
    "t_timestamp",
    "t_uint8",
    "t_uint16",
    "t_uint32",
    "t_uint64",
    "t_uint128",
    "t_unsigned_integer",
    "t_utf8_char",
    "t_utf8_string",
]



"""Module for controle flow AST."""

from __future__ import annotations

from typing import Optional, cast

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DictDataTypesStruct,
    Expr,
    Identifier,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.blocks import Block
from astx.tools.typing import typechecked
from astx.variables import InlineVariableDeclaration


@public
@typechecked
class IfStmt(StatementType):
    """AST class for `if` statement."""

    condition: Expr
    then: Block
    else_: Optional[Block]

    def __init__(
        self,
        condition: Expr,
        then: Block,
        else_: Optional[Block] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the IfStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.loc = loc
        self.condition = condition
        self.then = then
        self.else_ = else_
        self.kind = ASTKind.IfStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"IfStmt[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        if_condition = {"condition": self.condition.get_struct(simplified)}
        if_then = {"then-block": self.then.get_struct(simplified)}
        if_else: ReprStruct = {}

        if self.else_ is not None:
            if_else = {"else-block": self.else_.get_struct(simplified)}

        key = f"IF-STMT[{id(self)}]" if simplified else "IF-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, if_condition),
            **cast(DictDataTypesStruct, if_then),
            **cast(DictDataTypesStruct, if_else),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class IfExpr(Expr):
    """AST class for `if` expression."""

    condition: Expr
    then: Block
    else_: Optional[Block]

    def __init__(
        self,
        condition: Expr,
        then: Block,
        else_: Optional[Block] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the IfExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.loc = loc
        self.condition = condition
        self.then = then
        self.else_ = else_
        self.kind = ASTKind.IfExprKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"IfExpr[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        if_condition = {"condition": self.condition.get_struct(simplified)}
        if_then = {"then-block": self.then.get_struct(simplified)}
        if_else: ReprStruct = {}

        if self.else_ is not None:
            if_else = {"else-block": self.else_.get_struct(simplified)}

        key = "IF-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, if_condition),
            **cast(DictDataTypesStruct, if_then),
            **cast(DictDataTypesStruct, if_else),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForRangeLoopStmt(StatementType):
    """AST class for `For` Range Statement."""

    variable: InlineVariableDeclaration
    start: Expr
    end: Expr
    step: Expr
    body: Block

    def __init__(
        self,
        variable: InlineVariableDeclaration,
        start: Expr,
        end: Expr,
        step: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForRangeLoopStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.variable = variable
        self.start = start
        self.end = end
        self.step = step
        self.body = body
        self.kind = ASTKind.ForRangeLoopStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        start = self.start
        end = self.end
        step = self.step
        var_name = self.variable.name
        return f"ForRangeLoopStmt({var_name}=[{start}:{end}:{step}])"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_start = {"start": self.start.get_struct(simplified)}
        for_end = {"end": self.end.get_struct(simplified)}
        for_step = {"step": self.step.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = (
            f"FOR-RANGE-LOOP-STMT[{id(self)}]"
            if simplified
            else "FOR-RANGE-LOOP-STMT"
        )
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_start),
            **cast(DictDataTypesStruct, for_end),
            **cast(DictDataTypesStruct, for_step),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForRangeLoopExpr(Expr):
    """AST class for `For` Range Expression."""

    variable: InlineVariableDeclaration
    start: Expr
    end: Expr
    step: Expr
    body: Block

    def __init__(
        self,
        variable: InlineVariableDeclaration,
        start: Expr,
        end: Expr,
        step: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForRangeLoopExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.variable = variable
        self.start = start
        self.end = end
        self.step = step
        self.body = body
        self.kind = ASTKind.ForRangeLoopExprKind
        # self.step = step if step is not None else LiteralInt32(1)

    def __str__(self) -> str:
        """Return a string that represents the object."""
        var_name = self.variable.name
        # note: it would be nice to have the following structure
        #    ForRangeLoopExpr({var_name}=[{start}:{end}:{step}])
        #    but we would need to have first something like a resolver
        #    otherwise it could be a very large output
        return f"ForRangeLoopExpr[{var_name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_var = {"var": self.variable.get_struct(simplified)}
        for_start = {"start": self.start.get_struct(simplified)}
        for_end = {"end": self.end.get_struct(simplified)}
        for_step = {"step": self.step.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = "FOR-RANGE-LOOP-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_var),
            **cast(DictDataTypesStruct, for_start),
            **cast(DictDataTypesStruct, for_end),
            **cast(DictDataTypesStruct, for_step),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForCountLoopStmt(StatementType):
    """
    AST class for a simple Count-Controlled `For` Loop statement.

    This is a very basic `for` loop, used by languages like C or C++.
    """

    initializer: InlineVariableDeclaration
    condition: Expr
    update: Expr
    body: Block

    def __init__(
        self,
        initializer: InlineVariableDeclaration,
        condition: Expr,
        update: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForCountLoopStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.initializer = initializer
        self.condition = condition
        self.update = update
        self.body = body
        self.kind = ASTKind.ForCountLoopStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        init = self.initializer
        cond = self.condition
        update = self.update
        return f"ForCountLoopStmt({init};{cond};{update})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_init = {"initialization": self.initializer.get_struct(simplified)}
        for_cond = {"condition": self.condition.get_struct(simplified)}
        for_update = {"update": self.update.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = f"FOR-COUNT-STMT[{id(self)}]" if simplified else "FOR-COUNT-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_init),
            **cast(DictDataTypesStruct, for_cond),
            **cast(DictDataTypesStruct, for_update),
            **cast(DictDataTypesStruct, for_body),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ForCountLoopExpr(Expr):
    """
    AST class for a simple Count-Controlled `For` Loop expression.

    This is a very basic `for` loop, used by languages like C or C++.
    """

    initializer: InlineVariableDeclaration
    condition: Expr
    update: Expr
    body: Block

    def __init__(
        self,
        initializer: InlineVariableDeclaration,
        condition: Expr,
        update: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ForLoopCountExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.initializer = initializer
        self.condition = condition
        self.update = update
        self.body = body
        self.kind = ASTKind.ForCountLoopExprKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        init = self.initializer
        cond = self.condition
        update = self.update
        return f"ForCountLoopExpr({init};{cond};{update})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_init = {"initialization": self.initializer.get_struct(simplified)}
        for_cond = {"condition": self.condition.get_struct(simplified)}
        for_update = {"update": self.update.get_struct(simplified)}
        for_body = self.body.get_struct(simplified)

        key = "FOR-COUNT-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_init),
            **cast(DictDataTypesStruct, for_cond),
            **cast(DictDataTypesStruct, for_update),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class AsyncForRangeLoopStmt(StatementType):
    """AST class for asynchronous `For` Range Statement."""

    variable: InlineVariableDeclaration
    start: Optional[Expr]
    end: Expr
    step: Optional[Expr]
    body: Block

    def __init__(
        self,
        variable: InlineVariableDeclaration,
        start: Optional[Expr],
        end: Expr,
        step: Optional[Expr],
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the AsyncForRangeLoopStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.variable = variable
        self.start = start
        self.end = end
        self.step = step
        self.body = body
        self.kind = ASTKind.AsyncRangeLoopStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        start = self.start
        end = self.end
        step = self.step
        var_name = self.variable.name
        return f"AsyncForRangeLoopStmt({var_name}=[{start}:{end}:{step}])"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_start = {
            "start": {}
            if self.start is None
            else self.start.get_struct(simplified)
        }
        for_end = {"end": self.end.get_struct(simplified)}
        for_step = {
            "step": {}
            if self.step is None
            else self.step.get_struct(simplified)
        }
        for_body = self.body.get_struct(simplified)

        key = "ASYNC-FOR-RANGE-LOOP-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_start),
            **cast(DictDataTypesStruct, for_end),
            **cast(DictDataTypesStruct, for_step),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class AsyncForRangeLoopExpr(Expr):
    """AST class for asynchronous `For` Range Expression."""

    variable: InlineVariableDeclaration
    start: Optional[Expr]
    end: Expr
    step: Optional[Expr]
    body: Block

    def __init__(
        self,
        variable: InlineVariableDeclaration,
        start: Optional[Expr],
        end: Expr,
        step: Optional[Expr],
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the AsyncForRangeLoopExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.variable = variable
        self.start = start
        self.end = end
        self.step = step
        self.body = body
        self.kind = ASTKind.AsyncRangeLoopExprKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        var_name = self.variable.name
        return f"AsyncForRangeLoopExpr[{var_name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        for_var = {"var": self.variable.get_struct(simplified)}
        for_start = {
            "start": {}
            if self.start is None
            else self.start.get_struct(simplified)
        }
        for_end = {"end": self.end.get_struct(simplified)}
        for_step = {
            "step": {}
            if self.step is None
            else self.step.get_struct(simplified)
        }
        for_body = self.body.get_struct(simplified)

        key = "ASYNC-FOR-RANGE-LOOP-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, for_var),
            **cast(DictDataTypesStruct, for_start),
            **cast(DictDataTypesStruct, for_end),
            **cast(DictDataTypesStruct, for_step),
            **cast(DictDataTypesStruct, for_body),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class BreakStmt(StatementType):
    """AST class for break statement."""

    def __init__(
        self,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the BreakStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.kind = ASTKind.BreakStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return "BreakStmt"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = "BREAK-STMT"
        value: DictDataTypesStruct = {}
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ContinueStmt(StatementType):
    """AST class for continue statement."""

    def __init__(
        self,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the ContinueStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.kind = ASTKind.ContinueStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return "ContinueStmt"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = "CONTINUE-STMT"
        value: DictDataTypesStruct = {}
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class WhileStmt(StatementType):
    """AST class for `while` statement."""

    condition: Expr
    body: Block

    def __init__(
        self,
        condition: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the WhileStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.condition = condition
        self.body = body
        self.kind = ASTKind.WhileStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"WhileStmt[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        while_condition = self.condition.get_struct(simplified)
        while_body = self.body.get_struct(simplified)

        key = f"WHILE-STMT[{id(self)}]" if simplified else "WHILE-STMT"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, while_condition),
            **cast(DictDataTypesStruct, while_body),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class WhileExpr(Expr):
    """AST class for `while` expression."""

    condition: Expr
    body: Block

    def __init__(
        self,
        condition: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the WhileExpr instance."""
        super().__init__(loc=loc, parent=parent)
        self.condition = condition
        self.body = body
        self.kind = ASTKind.WhileExprKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"WhileExpr[{self.condition}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        while_condition = self.condition.get_struct(simplified)
        while_body = self.body.get_struct(simplified)

        key = "WHILE-EXPR"
        value: ReprStruct = {
            **cast(DictDataTypesStruct, while_condition),
            **cast(DictDataTypesStruct, while_body),
        }

        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class CaseStmt(StatementType):
    """AST class for a case in a Switch statement."""

    condition: Optional[Expr] = None
    body: Block
    default: bool = False

    def __init__(
        self,
        body: Block,
        condition: Optional[Expr] = None,
        default: bool = False,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the CaseStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.condition = condition
        self.body = body
        self.default = default
        self.kind = ASTKind.CaseStmtKind

        if self.default is False and self.condition is None:
            raise ValueError(
                "Condition must be provided for non-default branches."
            )

        if self.default is True and self.condition is not None:
            raise ValueError(
                "Condition must NOT be provided for default branches."
            )

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return (
            f"CaseStmt[{self.condition}]"
            if self.condition
            else "CaseStmt[default]"
        )

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        default_case = "default" if self.condition is None else ""
        default_only = "[default]" if self.condition is None else ""
        id_str = f"{id(self)}" if simplified else ""

        key = (
            f"CASE-STMT[{id_str}{default_case}]"
            if simplified and self.condition is not None
            else f"CASE-STMT[{id_str}, {default_case}]"
            if simplified
            else f"CASE-STMT{default_only}"
        )

        condition_dict = (
            {}
            if self.condition is None
            else {"condition": self.condition.get_struct(simplified)}
        )
        value = {
            **cast(DictDataTypesStruct, condition_dict),
            "body": self.body.get_struct(simplified),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class SwitchStmt(StatementType):
    """AST class for Switch statements based on Rust's match syntax."""

    value: Expr
    cases: ASTNodes[CaseStmt]

    def __init__(
        self,
        value: Expr,
        cases: list[CaseStmt] | ASTNodes[CaseStmt],
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the SwitchStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.value = value

        if isinstance(cases, ASTNodes):
            self.cases = cases
        else:
            self.cases = ASTNodes[CaseStmt]()
            for case in cases:
                self.cases.append(case)

        self.kind = ASTKind.SwitchStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"SwitchStmt[{len(self.cases)}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"SWITCH-STMT[{id(self)}]" if simplified else "SWITCH-STMT"
        case_dict = {}
        for d in range(len(self.cases)):
            case_dict[f"case_{d}"] = self.cases[d].get_struct(simplified)

        value: DictDataTypesStruct = {
            "value": self.value.get_struct(simplified),
            **cast(DictDataTypesStruct, {"cases": case_dict}),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class GotoStmt(StatementType):
    """AST class for function `Goto` statement."""

    label: Identifier

    def __init__(
        self,
        label: Identifier,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the Return instance."""
        super().__init__(loc=loc, parent=parent)
        self.label = label
        self.kind = ASTKind.GotoStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"Goto[{self.label.value}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"GOTO-STMT[{self.label.value}]"
        value: DictDataTypesStruct = {}
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class DoWhileStmt(WhileStmt):
    """AST class for `do-while` statement."""

    def __init__(
        self,
        condition: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the DoWhileStmt instance."""
        super().__init__(
            condition=condition, body=body, loc=loc, parent=parent
        )
        self.kind = ASTKind.DoWhileStmtKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"DoWhileStmt[{self.condition}]"


@public
@typechecked
class DoWhileExpr(WhileExpr):
    """AST class for `do-while` expression."""

    def __init__(
        self,
        condition: Expr,
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the DoWhileExpr instance."""
        super().__init__(
            condition=condition, body=body, loc=loc, parent=parent
        )
        self.kind = ASTKind.DoWhileExprKind

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"DoWhileExpr[{self.condition}]"



"""Module for subscripts definitions/declarations."""

from typing import Optional, cast

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DictDataTypesStruct,
    Expr,
    ReprStruct,
    SourceLocation,
)
from astx.literals import LiteralNone
from astx.tools.typing import typechecked


@public
@typechecked
class SubscriptExpr(Expr):
    """AST class for subscript expressions."""

    value: Expr
    index: Expr
    lower: Expr
    upper: Expr
    step: Expr

    def __init__(
        self,
        value: Expr,
        index: Optional[Expr] = None,
        lower: Optional[Expr] = None,
        upper: Optional[Expr] = None,
        step: Optional[Expr] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """
        Initialize the SubscriptExpr instance.

        Parameters
        ----------
        value: Expr
            The expression representing the object being indexed (e.g.,
        an array or list).
        index (optional): Expr
            The index of the variable.
        lower (optional): Expr
            The lower bound of the slice (inclusive).
        upper (optional): Expr
            The upper bound of the slice (exclusive).
        step (optional): Expr
            The step size for the slice.
        loc: SourceLocation
            The source location of the expression.
        parent (optional): ASTNodes
            The parent AST node.
        """
        super().__init__(loc=loc, parent=parent)
        self.value: Expr = value if value is not None else LiteralNone()
        self.index: Expr = index if index is not None else LiteralNone()
        self.lower: Expr = lower if lower is not None else LiteralNone()
        self.upper: Expr = upper if upper is not None else LiteralNone()
        self.step: Expr = step if step is not None else LiteralNone()
        self.kind = ASTKind.SubscriptExprKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        lower_str = (
            str(self.lower)
            if not isinstance(self.lower, LiteralNone)
            else str(self.index)
        )

        upper_str = (
            ":" + str(self.upper)
            if not isinstance(self.upper, LiteralNone)
            else ""
        )
        step_str = (
            ":" + str(self.step)
            if not isinstance(self.step, LiteralNone)
            else ""
        )
        return f"SubscriptExpr({self.value}[{lower_str}{upper_str}{step_str}])"

    def _get_struct_wrapper(self, simplified: bool) -> DictDataTypesStruct:
        """Return the AST structure of the object."""
        value_dict: ReprStruct = {"indexed": self.value.get_struct(simplified)}

        lower_key = "index" if isinstance(self.lower, LiteralNone) else "lower"
        lower_value = (
            self.index.get_struct(simplified)
            if isinstance(self.lower, LiteralNone)
            else self.lower.get_struct(simplified)
        )
        lower_dict: ReprStruct = {lower_key: lower_value}

        upper_dict: ReprStruct = {}
        if not isinstance(self.upper, LiteralNone):
            upper_dict = {"upper": self.upper.get_struct(simplified)}

        step_dict: ReprStruct = {}
        if not isinstance(self.step, LiteralNone):
            step_dict = {"step": self.step.get_struct(simplified)}

        value: DictDataTypesStruct = {
            **cast(DictDataTypesStruct, value_dict),
            **cast(DictDataTypesStruct, lower_dict),
            **cast(DictDataTypesStruct, upper_dict),
            **cast(DictDataTypesStruct, step_dict),
        }
        return value

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = "SubscriptExpr"
        value = self._get_struct_wrapper(simplified)

        return self._prepare_struct(key, value, simplified)



"""Collection of ASTx nodes used for types."""

from astx.types.base import (
    AnyType,
    NoneType,
)
from astx.types.boolean import (
    Boolean,
)
from astx.types.casting import (
    TypeCastExpr,
)
from astx.types.collections import (
    CollectionType,
    DictType,
    ListType,
    SetType,
    TupleType,
)
from astx.types.numeric import (
    Complex,
    Complex32,
    Complex64,
    Float16,
    Float32,
    Float64,
    Floating,
    Int8,
    Int16,
    Int32,
    Int64,
    Integer,
    Number,
    SignedInteger,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    UInt128,
    UnsignedInteger,
)
from astx.types.operators import (
    AndOp,
    BinaryOp,
    BoolBinaryOp,
    BoolUnaryOp,
    DataTypeOps,
    NandOp,
    NorOp,
    NotOp,
    OrOp,
    UnaryOp,
    XnorOp,
    XorOp,
)
from astx.types.string import (
    String,
    UTF8Char,
    UTF8String,
)
from astx.types.temporal import (
    Date,
    DateTime,
    Time,
    Timestamp,
)

__all__ = [
    "AndOp",
    "AnyType",
    "BinaryOp",
    "BoolBinaryOp",
    "BoolUnaryOp",
    "Boolean",
    "CollectionType",
    "Complex",
    "Complex32",
    "Complex64",
    "DataTypeOps",
    "Date",
    "DateTime",
    "DictType",
    "Float16",
    "Float32",
    "Float64",
    "Floating",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Integer",
    "ListType",
    "NandOp",
    "NoneType",
    "NorOp",
    "NotOp",
    "Number",
    "OrOp",
    "SetType",
    "SignedInteger",
    "String",
    "Time",
    "Timestamp",
    "TupleType",
    "TypeCastExpr",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "UInt128",
    "UTF8Char",
    "UTF8String",
    "UnaryOp",
    "UnsignedInteger",
    "XnorOp",
    "XorOp",
]



"""ASTx class for With Statement (Context Manager)."""

from __future__ import annotations

from typing import Dict, List, Optional, cast

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DataTypesStruct,
    Expr,
    Identifier,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.blocks import Block


@public
class WithItem:
    """AST class representing an item inside a `with` statement."""

    def __init__(
        self, context_expr: Expr, instance_name: Optional[Identifier] = None
    ) -> None:
        """Initialize a WithItem instance."""
        self.context_expr = context_expr
        self.instance_name = instance_name

    def __str__(self) -> str:
        """Return string representation of the WithItem."""
        if self.instance_name:
            return f"{self.context_expr} as {self.instance_name}"
        return str(self.context_expr)

    def _prepare_struct(
        self, key: str, value: DataTypesStruct, simplified: bool = False
    ) -> Dict[str, DataTypesStruct]:
        """Prepare structural representation."""
        return {key: value} if simplified else {"WithItem": {key: value}}

    def get_struct(
        self, simplified: bool = False
    ) -> Dict[str, DataTypesStruct]:
        """Get structural representation of the WithItem."""
        key = (
            "CONTEXT"
            if not self.instance_name
            else f"CONTEXT[{self.context_expr!s}]"
        )
        value = cast(DataTypesStruct, self.context_expr.get_struct(simplified))
        return self._prepare_struct(key, value, simplified)


class WithStmt(StatementType):
    """AST class for the `with` statement (context manager)."""

    def __init__(
        self,
        items: List[WithItem],
        body: Block,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize WithStmt instance."""
        super().__init__(loc=loc, parent=parent)
        self.items = items
        self.body = body
        self.kind = ASTKind.WithStmtKind

    def __str__(self) -> str:
        """Return string representation of the WithStmt."""
        items_str = ", ".join(str(item) for item in self.items)
        return f"WithStmt[{items_str}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Get structural representation of the WithStmt."""
        items_structs: List[Dict[str, DataTypesStruct]] = [
            item.get_struct(simplified) for item in self.items
        ]

        return cast(
            ReprStruct,
            {
                "WITH-STMT": {
                    "items": items_structs,
                    "body": self.body.get_struct(simplified),
                }
            },
        )



"""ASTx Collection Data Types."""

from __future__ import annotations

from typing import List

from public import public

from astx.base import ExprType
from astx.tools.typing import typechecked
from astx.types.base import AnyType


@public
@typechecked
class CollectionType(AnyType):
    """Base class for collection data types."""


@public
@typechecked
class ListType(CollectionType):
    """List data type expression."""

    def __init__(self, element_types: List[ExprType]) -> None:
        """Initialize ListType with an element type."""
        self.element_types = element_types

    def __str__(self) -> str:
        """Return string representation of ListType."""
        types_str = ", ".join(str(t) for t in self.element_types)
        return f"ListType[{types_str}]"


@public
@typechecked
class SetType(CollectionType):
    """Set data type expression."""

    def __init__(self, element_type: ExprType) -> None:
        """Initialize SetType with an element type."""
        self.element_type = element_type

    def __str__(self) -> str:
        """Return string representation of SetType."""
        return f"SetType[{self.element_type}]"


@public
@typechecked
class DictType(CollectionType):
    """Dictionary data type expression."""

    def __init__(self, key_type: ExprType, value_type: ExprType) -> None:
        """Initialize DictType with key-value types."""
        self.key_type = key_type
        self.value_type = value_type

    def __str__(self) -> str:
        """Return string representation of DictType."""
        return f"DictType[{self.key_type}, {self.value_type}]"


@public
@typechecked
class TupleType(CollectionType):
    """Tuple data type expression."""

    def __init__(self, element_types: List[ExprType]) -> None:
        """Initialize TupleType with multiple element types."""
        self.element_types = element_types

    def __str__(self) -> str:
        """Return string representation of TupleType."""
        types_str = ", ".join(str(t) for t in self.element_types)
        return f"TupleType[{types_str}]"



"""ASTx Collection Literals."""

from __future__ import annotations

from typing import Dict, List, Set, Tuple

from public import public

from astx.base import NO_SOURCE_LOCATION, SourceLocation
from astx.literals.base import Literal
from astx.tools.typing import typechecked
from astx.types.collections import DictType, ListType, SetType, TupleType
from astx.types.numeric import Int32


@public
@typechecked
class LiteralList(Literal):
    """Literal representation of a List."""

    elements: List[Literal]

    def __init__(
        self, elements: List[Literal], loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralList."""
        super().__init__(loc)
        self.elements = list(elements)  # Ensure correct type
        unique_types = {type(elem.type_) for elem in elements}
        self.type_ = ListType([t() for t in unique_types])
        self.loc = loc


@public
@typechecked
class LiteralTuple(Literal):
    """Literal representation of a Tuple."""

    elements: Tuple[Literal, ...]

    def __init__(
        self,
        elements: Tuple[Literal, ...],
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize LiteralTuple."""
        super().__init__(loc)
        self.elements = elements
        self.type_ = TupleType([elem.type_ for elem in elements])
        self.loc = loc


@public
@typechecked
class LiteralSet(Literal):
    """Literal representation of a Set."""

    elements: Set[Literal]

    def __init__(
        self, elements: Set[Literal], loc: SourceLocation = NO_SOURCE_LOCATION
    ) -> None:
        """Initialize LiteralSet."""
        super().__init__(loc)
        self.elements = set(elements)
        unique_types = {type(elem.type_) for elem in elements}
        self.type_ = SetType(
            unique_types.pop()() if len(unique_types) == 1 else Int32()
        )
        self.loc = loc


@public
@typechecked
class LiteralDict(Literal):
    """Literal representation of a Dictionary."""

    elements: Dict[Literal, Literal]

    def __init__(
        self,
        elements: Dict[Literal, Literal],
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize LiteralDict."""
        super().__init__(loc)
        self.elements = dict(elements)
        key_types = {type(k.type_) for k in elements.keys()}
        value_types = {type(v.type_) for v in elements.values()}
        self.type_ = DictType(
            key_types.pop()() if len(key_types) == 1 else Int32(),
            value_types.pop()() if len(value_types) == 1 else Int32(),
        )
        self.loc = loc



"""ASTx Data Types module."""

from __future__ import annotations

from typing import Optional

from public import public

from astx.base import (
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DataType,
    ExprType,
    ReprStruct,
    SourceLocation,
)
from astx.tools.typing import typechecked


@public
@typechecked
class DataTypeOps(DataType):
    """Overload some magic functions used for the main operations."""

    def __hash__(self) -> int:
        """Ensure that the hash method is not None."""
        return super().__hash__()

    def __add__(self, other: DataType) -> BinaryOp:
        """Overload the magic `add` method."""
        return BinaryOp("+", self, other)

    def __eq__(self, other: DataType) -> BinaryOp:  # type: ignore
        """Overload the magic `eq` method."""
        return BinaryOp("==", self, other)

    def __floordiv__(self, other: DataType) -> BinaryOp:
        """Overload the magic `floordiv` method."""
        return BinaryOp("//", self, other)

    def __ge__(self, other: DataType) -> BinaryOp:
        """Overload the magic `ge` method."""
        return BinaryOp(">=", self, other)

    def __gt__(self, other: DataType) -> BinaryOp:
        """Overload the magic `gt` method."""
        return BinaryOp(">", self, other)

    def __le__(self, other: DataType) -> BinaryOp:
        """Overload the magic `le` method."""
        return BinaryOp("<=", self, other)

    def __lt__(self, other: DataType) -> BinaryOp:
        """Overload the magic `lt` method."""
        return BinaryOp("<", self, other)

    def __mod__(self, other: DataType) -> BinaryOp:
        """Overload the magic `mod` method."""
        return BinaryOp("%", self, other)

    def __mul__(self, other: DataType) -> BinaryOp:
        """Overload the magic `mul` method."""
        return BinaryOp("*", self, other)

    def __ne__(self, other: DataType) -> BinaryOp:  # type: ignore
        """Overload the magic `ne` method."""
        return BinaryOp("!=", self, other)

    def __neg__(self) -> UnaryOp:
        """Overload the magic `neg` method."""
        return UnaryOp("-", self)

    def __pos__(self) -> UnaryOp:
        """Overload the magic `pos` method."""
        return UnaryOp("+", self)

    def __pow__(self, other: DataType) -> BinaryOp:
        """Overload the magic `pow` method."""
        return BinaryOp("^", self, other)

    def __sub__(self, other: DataType) -> BinaryOp:
        """Overload the magic `sub` method."""
        return BinaryOp("-", self, other)

    def __truediv__(self, other: DataType) -> BinaryOp:
        """Overload the magic `truediv` method."""
        return BinaryOp("/", self, other)

    def __and__(self, other: DataType) -> AndOp:
        """Overload the magic 'and' method."""
        return AndOp(self, other)

    def __or__(self, other: DataType) -> OrOp:
        """Overload the magic 'or' method."""
        return OrOp(self, other)

    def __xor__(self, other: DataType) -> XorOp:
        """Overload the magic 'xor' method."""
        return XorOp(self, other)

    def __invert__(self) -> NotOp:
        """Overload the magic 'not' method."""
        return NotOp(self)


@public
@typechecked
class UnaryOp(DataTypeOps):
    """AST class for the unary operator."""

    op_code: str
    operand: DataType

    def __init__(
        self,
        op_code: str,
        operand: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the UnaryOp instance."""
        super().__init__(loc=loc, parent=parent)
        self.op_code = op_code
        self.operand = operand
        self.kind = ASTKind.UnaryOpKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"UnaryOp[{self.op_code}]({self.operand})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"UNARY[{self.op_code}]"
        value = self.operand.get_struct(simplified)
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class BinaryOp(DataTypeOps):
    """AST class for the binary operator."""

    type_: ExprType
    lhs: DataType
    rhs: DataType
    op_code: str

    def __init__(
        self,
        op_code: str,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Initialize the BinaryOp instance."""
        super().__init__(loc=loc, parent=parent)

        self.op_code = op_code
        self.lhs = lhs
        self.rhs = rhs
        self.kind = ASTKind.BinaryOpKind

        if not (
            isinstance(lhs.type_, DataType) and isinstance(rhs.type_, DataType)
        ):
            raise Exception(
                "For now, binary operators are just allowed for `DataType`."
                f"LHS: {lhs.type_}, RHS: {rhs.type_}"
            )

        if lhs.type_ == rhs.type_:
            self.type_ = lhs.type_
        else:
            # type inference
            self.type_ = max([lhs.type_, rhs.type_], key=lambda v: v.nbytes)

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"BinaryOp[{self.op_code}]({self.lhs},{self.rhs})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure that represents the object."""
        key = f"BINARY[{self.op_code}]"
        lhs = {"lhs": self.lhs.get_struct(simplified)}
        rhs = {"rhs": self.rhs.get_struct(simplified)}

        content: ReprStruct = {**lhs, **rhs}
        return self._prepare_struct(key, content, simplified)


@public
@typechecked
class BoolBinaryOp(BinaryOp):
    """Base AST class for boolean binary operations."""

    def __init__(
        self,
        op_code: str,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(
            op_code=op_code,
            lhs=lhs,
            rhs=rhs,
            loc=loc,
            parent=parent,
        )

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"({self.lhs} {self.op_code} {self.rhs})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure that represents the object."""
        key = f"BOOL_BINARY_OP[{self.__class__.__name__}]"
        value: ReprStruct = {
            "lhs": self.lhs.get_struct(simplified),
            "rhs": self.rhs.get_struct(simplified),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class BoolUnaryOp(UnaryOp):
    """Base AST class for boolean unary operations."""

    def __init__(
        self,
        op_code: str,
        operand: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        super().__init__(
            op_code=op_code,
            operand=operand,
            loc=loc,
            parent=parent,
        )

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"({self.op_code} {self.operand})"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure that represents the object."""
        key = f"BOOL_UNARY_OP[{self.__class__.__name__}]"
        value: ReprStruct = {"operand": self.operand.get_struct(simplified)}
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class AndOp(BoolBinaryOp):
    """AST class for logical AND operation."""

    kind = ASTKind.AndOpKind
    op_code = "and"

    def __init__(
        self,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Instantiate AST class for logical AND operation."""
        super().__init__(
            op_code=self.op_code,
            lhs=lhs,
            rhs=rhs,
            loc=loc,
            parent=parent,
        )


@public
@typechecked
class OrOp(BoolBinaryOp):
    """AST class for logical OR operation."""

    kind = ASTKind.OrOpKind
    op_code = "or"

    def __init__(
        self,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Instantiate AST class for logical OR operation."""
        super().__init__(
            op_code=self.op_code,
            lhs=lhs,
            rhs=rhs,
            loc=loc,
            parent=parent,
        )


@public
@typechecked
class XorOp(BoolBinaryOp):
    """AST class for logical XOR operation."""

    kind = ASTKind.XorOpKind
    op_code = "xor"

    def __init__(
        self,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Instantiate AST class for logical XOR operation."""
        super().__init__(
            op_code=self.op_code,
            lhs=lhs,
            rhs=rhs,
            loc=loc,
            parent=parent,
        )


@public
@typechecked
class NandOp(BoolBinaryOp):
    """AST class for logical NAND operation."""

    kind = ASTKind.NandOpKind
    op_code = "nand"

    def __init__(
        self,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Instantiate AST class for logical NAND operation."""
        super().__init__(
            op_code=self.op_code,
            lhs=lhs,
            rhs=rhs,
            loc=loc,
            parent=parent,
        )


@public
@typechecked
class NorOp(BoolBinaryOp):
    """AST class for logical NOR operation."""

    kind = ASTKind.NorOpKind
    op_code = "nor"

    def __init__(
        self,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Instantiate AST class for logical NOR operation."""
        super().__init__(
            op_code=self.op_code,
            lhs=lhs,
            rhs=rhs,
            loc=loc,
            parent=parent,
        )


@public
@typechecked
class XnorOp(BoolBinaryOp):
    """AST class for logical XNOR operation."""

    kind = ASTKind.XnorOpKind
    op_code = "xnor"

    def __init__(
        self,
        lhs: DataType,
        rhs: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Instantiate AST class for logical XNOR operation."""
        super().__init__(
            op_code=self.op_code,
            lhs=lhs,
            rhs=rhs,
            loc=loc,
            parent=parent,
        )


@public
@typechecked
class NotOp(BoolUnaryOp):
    """AST class for logical NOT operation."""

    kind = ASTKind.NotOpKind
    op_code = "not"

    def __init__(
        self,
        operand: DataType,
        loc: SourceLocation = NO_SOURCE_LOCATION,
        parent: Optional[ASTNodes] = None,
    ) -> None:
        """Instantiate AST class for logical NOT operation."""
        super().__init__(
            op_code=self.op_code,
            operand=operand,
            loc=loc,
            parent=parent,
        )



"""Module for Exceptions."""

from __future__ import annotations

from typing import Iterable, Optional, cast

from public import public

from astx.base import (
    AST,
    NO_SOURCE_LOCATION,
    ASTKind,
    ASTNodes,
    DictDataTypesStruct,
    Expr,
    Identifier,
    ReprStruct,
    SourceLocation,
    StatementType,
)
from astx.blocks import Block
from astx.tools.typing import typechecked


@public
@typechecked
class ThrowStmt(StatementType):
    """AST class for throw statements."""

    exception: Optional[Expr]

    def __init__(
        self,
        exception: Optional[Expr] = None,
        parent: Optional[ASTNodes] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the instance."""
        super().__init__(loc=loc, parent=parent)
        self.exception = exception
        self.kind = ASTKind.ThrowStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        throw_str = (
            f"ThrowStmt[{self.exception}]" if self.exception else "ThrowStmt"
        )
        return throw_str

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"THROW-STMT[{id(self)}]" if simplified else "THROW-STMT"
        value = self.exception.get_struct(simplified) if self.exception else ""
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class CatchHandlerStmt(StatementType):
    """AST class for catch statements."""

    body: Block[AST]
    name: Optional[Identifier]
    types: Optional[ASTNodes[Identifier]]

    def __init__(
        self,
        body: Block[AST],
        name: Optional[Identifier] = None,
        types: Optional[Iterable[Identifier] | ASTNodes[Identifier]] = None,
        parent: Optional[ASTNodes] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the instance."""
        super().__init__(loc=loc, parent=parent)
        self.body = body
        self.name = name

        if types:
            if isinstance(types, ASTNodes):
                self.types = types
            else:
                self.types = ASTNodes[Identifier]()
                for t in types:
                    self.types.append(t)
        else:
            self.types = None

        self.kind = ASTKind.CatchHandlerStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return f"CatchHandlerStmt[{self.name}]"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = (
            f"CATCH-HANDLER-STMT[{id(self)}]"
            if simplified
            else "CATCH-HANDLER-STMT"
        )
        body_dict = {"body": self.body.get_struct(simplified)}
        name_dict = (
            {"name": self.name.get_struct(simplified)} if self.name else {}
        )
        types_dict = (
            {"types": self.types.get_struct(simplified)} if self.types else {}
        )

        value: DictDataTypesStruct = {
            **cast(DictDataTypesStruct, body_dict),
            **cast(DictDataTypesStruct, name_dict),
            **cast(DictDataTypesStruct, types_dict),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class ExceptionHandlerStmt(StatementType):
    """AST class for try statements."""

    body: Block[AST]
    handlers: ASTNodes[CatchHandlerStmt]
    finally_handler: Optional[FinallyHandlerStmt]

    def __init__(
        self,
        body: Block[AST],
        handlers: Iterable[CatchHandlerStmt] | ASTNodes[CatchHandlerStmt],
        finally_handler: Optional[FinallyHandlerStmt] = None,
        parent: Optional[ASTNodes] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the instance."""
        super().__init__(loc=loc, parent=parent)
        self.body = body

        if isinstance(handlers, ASTNodes):
            self.handlers = handlers
        else:
            self.handlers = ASTNodes[CatchHandlerStmt]()
            for h in handlers:
                self.handlers.append(h)

        self.finally_handler = finally_handler

        self.kind = ASTKind.ExceptionHandlerStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return "ExceptionHandlerStmt"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = (
            f"EXCEPTION-HANDLER-STMT[{id(self)}]"
            if simplified
            else "EXCEPTION-HANDLER-STMT"
        )

        body_dict = {"body": self.body.get_struct(simplified)}
        handlers_dict = {"handlers": self.handlers.get_struct(simplified)}
        finally_dict = (
            {"finally_handler": self.finally_handler.get_struct(simplified)}
            if self.finally_handler
            else {}
        )

        value: DictDataTypesStruct = {
            **cast(DictDataTypesStruct, body_dict),
            **cast(DictDataTypesStruct, handlers_dict),
            **cast(DictDataTypesStruct, finally_dict),
        }
        return self._prepare_struct(key, value, simplified)


@public
@typechecked
class FinallyHandlerStmt(StatementType):
    """AST class for finally statements."""

    body: Block[AST]

    def __init__(
        self,
        body: Block[AST],
        parent: Optional[ASTNodes] = None,
        loc: SourceLocation = NO_SOURCE_LOCATION,
    ) -> None:
        """Initialize the instance."""
        super().__init__(loc=loc, parent=parent)
        self.body = body
        self.kind = ASTKind.FinallyHandlerStmtKind

    def __str__(self) -> str:
        """Return a string that represents the object."""
        return "FinallyStmt"

    def get_struct(self, simplified: bool = False) -> ReprStruct:
        """Return the AST structure of the object."""
        key = f"FINALLY-STMT[{id(self)}]" if simplified else "FINALLY-STMT"
        value: DictDataTypesStruct = {"body": self.body.get_struct(simplified)}

        return self._prepare_struct(key, value, simplified)
